"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
// Native.
var events_1 = require("events");
// Package.
var debug = require("debug");
var error = require("../utils/error");
var Frame_1 = require("./Frame");
var JsHandle_1 = require("./JsHandle");
var ElementHandle_1 = require("./ElementHandle");
var ExecutionContext_1 = require("./ExecutionContext");
// Code.
var debugVerbose = debug('di:verbose:frameManager');
var debugWarn = debug('di:warn:frameManager');
var FrameManager = /** @class */ (function (_super) {
    __extends(FrameManager, _super);
    function FrameManager(client, frameTree, page) {
        var _this = _super.call(this) || this;
        _this.contextIdToContext = new Map();
        _this.frames = new Map();
        _this.client = client;
        _this.page = page;
        _this.client.on('Page.frameAttached', function (event) {
            return _this.onFrameAttached(event.frameId, event.parentFrameId);
        });
        _this.client.on('Page.frameNavigated', function (event) {
            return _this.onFrameNavigated(event.frame);
        });
        _this.client.on('Page.frameDetached', function (event) {
            return _this.onFrameDetached(event.frameId);
        });
        _this.client.on('Runtime.executionContextCreated', function (event) {
            return _this.onExecutionContextCreated(event.context);
        });
        _this.client.on('Runtime.executionContextDestroyed', function (event) {
            return _this.onExecutionContextDestroyed(event.executionContextId);
        });
        _this.client.on('Runtime.executionContextsCleared', function () {
            return _this.onExecutionContextsCleared();
        });
        _this.client.on('Page.lifecycleEvent', function (event) {
            return _this.onLifecycleEvent(event);
        });
        _this.handleFrameTree(frameTree);
        return _this;
    }
    Object.defineProperty(FrameManager, "Events", {
        get: function () {
            return {
                FrameAttached: 'frameattached',
                FrameNavigated: 'framenavigated',
                FrameDetached: 'framedetached',
                LifecycleEvent: 'lifecycleevent',
            };
        },
        enumerable: true,
        configurable: true
    });
    FrameManager.prototype.handleFrameTree = function (frameTree) {
        if (frameTree.frame.parentId) {
            this.onFrameAttached(frameTree.frame.id, frameTree.frame.parentId);
        }
        this.onFrameNavigated(frameTree.frame);
        if (!frameTree.childFrames) {
            return;
        }
        for (var _i = 0, _a = frameTree.childFrames; _i < _a.length; _i++) {
            var child = _a[_i];
            this.handleFrameTree(child);
        }
    };
    FrameManager.prototype.removeContext = function (context) {
        debugVerbose("removeContext: %j", { contextId: context.contextId });
        var frameId = context.frameId;
        if (frameId && this.frames.has(frameId) && context.isDefault) {
            var frame = this.frames.get(frameId);
            frame.setDefaultContext(undefined);
        }
    };
    FrameManager.prototype.onLifecycleEvent = function (event) {
        var frame = this.frames.get(event.frameId);
        if (frame) {
            frame.onLifecycleEvent(event.loaderId, event.name);
            this.emit(FrameManager.Events.LifecycleEvent, frame);
        }
    };
    FrameManager.prototype.onExecutionContextCreated = function (contextPayload) {
        debugVerbose("executionContextCreated %j", {
            contextId: contextPayload.id,
        });
        var context = new ExecutionContext_1.default(this.client, contextPayload, this.createJSHandle.bind(this, contextPayload.id));
        this.contextIdToContext.set(contextPayload.id, context);
        var frameId = context.frameId;
        if (frameId) {
            if (this.frames.has(frameId) && context.isDefault) {
                var frame = this.frames.get(frameId);
                frame.setDefaultContext(context);
            }
        }
    };
    FrameManager.prototype.onExecutionContextDestroyed = function (executionContextId) {
        debugVerbose("executionContextDestroyed %j", {
            executionContextId: executionContextId,
        });
        if (this.contextIdToContext.has(executionContextId)) {
            var context = this.contextIdToContext.get(executionContextId);
            this.contextIdToContext.delete(executionContextId);
            this.removeContext(context);
        }
    };
    FrameManager.prototype.onExecutionContextsCleared = function () {
        debugVerbose("executionContextCleared");
        for (var _i = 0, _a = Array.from(this.contextIdToContext.values()); _i < _a.length; _i++) {
            var context = _a[_i];
            this.removeContext(context);
        }
        this.contextIdToContext.clear();
    };
    FrameManager.prototype.onFrameAttached = function (frameId, parentFrameId) {
        debugVerbose("frameAttached %j", { frameId: frameId, parentFrameId: parentFrameId });
        // Do not register twice a same frame.
        if (this.frames.has(frameId)) {
            debugWarn("Frame already set: " + frameId);
            return;
        }
        // Track possible flow inconsistency.
        if (!this.frames.has(parentFrameId)) {
            throw new error.FlowError("FrameManager: Page.frameAttached seems to be called before Page.frameNavigated");
        }
        var parentFrame = this.frames.get(parentFrameId);
        var frame = new Frame_1.default(this.client, this.page, parentFrame, frameId);
        this.frames.set(frame.id, frame);
        this.emit(FrameManager.Events.FrameAttached, frame);
    };
    FrameManager.prototype.onFrameNavigated = function (framePayload) {
        var isMainFrame = !framePayload.parentId;
        debugVerbose("frameNavigated: %j", {
            frameId: framePayload.id,
            isMainFrame: isMainFrame,
        });
        var frame;
        if (isMainFrame) {
            frame = this.mainFrame;
        }
        else if (this.frames.has(framePayload.id)) {
            frame = this.frames.get(framePayload.id);
        }
        else {
            throw new error.FlowError("FrameManager: Page.frameNavigated has completed either on a top level frame or old version frame");
        }
        // Detach all child frames first.
        if (frame) {
            for (var _i = 0, _a = frame.getChildFrames(); _i < _a.length; _i++) {
                var child = _a[_i];
                this.removeFramesRecursively(child);
            }
        }
        // Update or create the main frame.
        if (isMainFrame) {
            if (frame) {
                debugVerbose("frameNavigated: update main frame");
                // Update frame id to retain frame identity on cross-process navigation.
                this.frames.delete(frame.id);
                frame.id = framePayload.id;
            }
            else {
                debugVerbose("frameNavigated: create main frame");
                // Initial main frame navigation.
                frame = new Frame_1.default(this.client, this.page, null, framePayload.id);
            }
            this.frames.set(framePayload.id, frame);
            this.mainFrame = frame;
        }
        // Update frame payload.
        if (frame) {
            debugVerbose("frameNavigated: update frame payload");
            frame.navigated(framePayload);
            this.emit(FrameManager.Events.FrameNavigated, frame);
        }
    };
    FrameManager.prototype.onFrameDetached = function (frameId) {
        debugVerbose("frameDetached: %j", { frameId: frameId });
        if (this.frames.has(frameId)) {
            var frame = this.frames.get(frameId);
            this.removeFramesRecursively(frame);
        }
    };
    FrameManager.prototype.removeFramesRecursively = function (frame) {
        debugVerbose("removeFramesRecursively of frame: %j", {
            frameId: frame.id,
        });
        for (var _i = 0, _a = frame.getChildFrames(); _i < _a.length; _i++) {
            var child = _a[_i];
            this.removeFramesRecursively(child);
        }
        frame.detach();
        this.frames.delete(frame.id);
        this.emit(FrameManager.Events.FrameDetached, frame);
    };
    FrameManager.prototype.createJSHandle = function (contextId, remoteObject) {
        if (this.contextIdToContext.has(contextId)) {
            var context = this.contextIdToContext.get(contextId);
            if (remoteObject.subtype === 'node') {
                return new ElementHandle_1.default(context, this.client, remoteObject, this.page);
            }
            return new JsHandle_1.default(context, this.client, remoteObject);
        }
        else {
            throw new error.FlowError("FrameManager: missing context with id: " + contextId);
        }
    };
    return FrameManager;
}(events_1.EventEmitter));
exports.default = FrameManager;
