"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
// Internal.
var Protocol = require("../types/protocol");
// Code.
var ProtocolCodeGenerator = /** @class */ (function () {
    function ProtocolCodeGenerator() {
        this.code = '';
        this.indent = '  ';
        this.indentStack = [''];
        this.refs = new Map();
    }
    Object.defineProperty(ProtocolCodeGenerator.prototype, "currentIndent", {
        get: function () {
            return this.indentStack[this.indentStack.length - 1];
        },
        enumerable: true,
        configurable: true
    });
    ProtocolCodeGenerator.prototype.generate = function (protocol) {
        var _this = this;
        this.code = '';
        this.indentStack.length = 1;
        this.appendProtocolVersionComment(protocol.version);
        var domains = new Set();
        each(protocol.domains, function (domain) {
            _this.gatherRefs(domain);
            domains.add(domain.domain);
        });
        each(protocol.domains, function (domain) {
            var domainName = domain.domain, events = domain.events, commands = domain.commands, types = domain.types;
            _this.append('');
            _this.appendComment(domain);
            _this.generateDomainTypeNamespace(domainName, function () {
                var refs = _this.getRefs(domainName);
                each(types, function (type) {
                    refs.set(type.id, true);
                    _this.appendComment(type);
                    _this.appendType(type);
                });
                _this.generatorMissingRefs(domainName, refs);
                each(events, function (event) {
                    _this.appendEventParametersType(event);
                });
                each(commands, function (command) {
                    _this.appendCommandTypes(command);
                });
            });
        });
        this.refs.forEach(function (refs, domainName) {
            if (!domains.has(domainName)) {
                _this.generateDomainTypeNamespace(domainName, function () {
                    _this.generatorMissingRefs(domainName, refs);
                });
            }
        });
        var code = this.code;
        this.code = undefined;
        this.refs.clear();
        return code;
    };
    ProtocolCodeGenerator.prototype.generatorMissingRefs = function (domainName, refs) {
        var _this = this;
        refs.forEach(function (value, key) {
            if (!value) {
                /* tslint:disable:no-console */
                console.warn("missing $ref " + domainName + "." + key);
                /* tslint:enable:no-console */
                _this.generateObjectTypeAlias(key, []);
            }
        });
    };
    ProtocolCodeGenerator.prototype.getRefs = function (domainName) {
        var refs = this.refs.get(domainName);
        if (!refs) {
            refs = new Map();
            this.refs.set(domainName, refs);
        }
        return refs;
    };
    ProtocolCodeGenerator.prototype.gatherRefs = function (domain) {
        var _this = this;
        var refs = this.getRefs(domain.domain);
        each(domain.types, function (type) {
            _this.gatherRefsFromDesc(type, refs);
        });
        each(domain.commands, function (command) {
            if (command.parameters) {
                _this.gatherRefsFromDescs(command.parameters, refs);
            }
            if (command.returns) {
                _this.gatherRefsFromDescs(command.returns, refs);
            }
        });
        each(domain.events, function (event) {
            if (event.parameters) {
                _this.gatherRefsFromDescs(event.parameters, refs);
            }
        });
        this.refs.set(domain.domain, refs);
    };
    ProtocolCodeGenerator.prototype.gatherRefsFromDescs = function (descs, refs) {
        for (var _i = 0, descs_1 = descs; _i < descs_1.length; _i++) {
            var desc = descs_1[_i];
            this.gatherRefsFromDesc(desc, refs);
        }
    };
    ProtocolCodeGenerator.prototype.gatherRefsFromDesc = function (desc, refs) {
        if (Protocol.isTypeRef(desc)) {
            var ref = desc.$ref;
            var period = ref.indexOf('.');
            if (period !== -1) {
                this.getRefs(ref.substring(0, period)).set(ref.substring(period + 1), false);
            }
            else {
                refs.set(ref, false);
            }
        }
        else if (Protocol.isArrayDescriptor(desc)) {
            this.gatherRefsFromDesc(desc.items, refs);
        }
        else if (Protocol.isObjectDescriptor(desc)) {
            if (desc.properties) {
                this.gatherRefsFromDescs(desc.properties, refs);
            }
        }
    };
    ProtocolCodeGenerator.prototype.appendProtocolVersionComment = function (version) {
        var versionString = '';
        if (version) {
            versionString = version.major + "." + version.minor + " ";
        }
        this.append('/**');
        this.append(" * Debugging Protocol " + versionString + "Domains");
        this.append(" * Generated on " + new Date());
        this.append(' */');
    };
    ProtocolCodeGenerator.prototype.generateDomainTypeNamespace = function (domainName, cb) {
        this.append("export namespace " + domainName + " {");
        this.block(cb);
        this.append('}');
    };
    ProtocolCodeGenerator.prototype.appendType = function (type) {
        var _this = this;
        if (Protocol.isTypeRef(type)) {
            this.append("export type " + type.id + " = " + type.$ref + ";");
        }
        else if (Protocol.isObjectDescriptor(type)) {
            var properties_1 = type.properties;
            if (properties_1 && properties_1.length > 0) {
                this.append("export interface " + type.id + " {");
                this.block(function () {
                    each(properties_1, function (prop) { return _this.generateProperty(prop); });
                });
                this.append('}');
            }
            else {
                this.append("export type " + type.id + " = any;");
            }
        }
        else {
            this.append("export type " + type.id + " = " + this.typeString(type) + ";");
        }
    };
    ProtocolCodeGenerator.prototype.appendEventParametersType = function (event) {
        if (event.parameters) {
            this.generateObjectTypeAlias("" + this.parametersTypeName(event.name), event.parameters);
        }
    };
    ProtocolCodeGenerator.prototype.block = function (cb) {
        this.indentStack.push(this.currentIndent + this.indent);
        cb();
        this.indentStack.pop();
    };
    ProtocolCodeGenerator.prototype.append = function (line) {
        this.code += this.currentIndent + line + '\n';
    };
    ProtocolCodeGenerator.prototype.generateProperty = function (desc) {
        this.appendComment(desc);
        this.append(this.namedTypeString(desc));
    };
    ProtocolCodeGenerator.prototype.appendComment = function (obj) {
        if (!obj.description) {
            return;
        }
        this.append("/** " + obj.description + " */");
    };
    ProtocolCodeGenerator.prototype.appendCommandTypes = function (command) {
        var name = command.name;
        if (command.parameters) {
            this.generateObjectTypeAlias(this.parametersTypeName(name), command.parameters);
        }
        if (command.returns) {
            this.generateObjectTypeAlias(this.returnTypeName(name), command.returns);
        }
    };
    ProtocolCodeGenerator.prototype.returnTypeName = function (name, domainName) {
        return buildTypeName(name, 'Return', domainName);
    };
    ProtocolCodeGenerator.prototype.parametersTypeName = function (name, domainName) {
        return buildTypeName(name, 'Parameters', domainName);
    };
    ProtocolCodeGenerator.prototype.generateObjectTypeAlias = function (name, props) {
        var _this = this;
        if (props && props.length) {
            this.append("export type " + name + " = {");
            this.block(function () {
                props.forEach(function (prop) { return _this.generateProperty(prop); });
            });
            this.append('};');
        }
        else {
            this.append("export type " + name + " = any;");
        }
    };
    ProtocolCodeGenerator.prototype.namedTypeString = function (desc) {
        return "" + desc.name + (desc.optional ? '?' : '') + ": " + this.typeString(desc) + ";";
    };
    ProtocolCodeGenerator.prototype.typeString = function (desc, isArray) {
        var _this = this;
        var typeName;
        var simple = true;
        if (Protocol.isTypeRef(desc)) {
            typeName = desc.$ref;
        }
        else if (Protocol.isObjectDescriptor(desc)) {
            var properties = desc.properties;
            if (properties && properties.length) {
                simple = false;
                typeName =
                    '{ ' + properties.map(function (p) { return _this.namedTypeString(p); }).join(' ') + ' }';
            }
            else {
                typeName = 'any';
            }
        }
        else if (Protocol.isArrayDescriptor(desc)) {
            typeName = this.typeString(desc.items, true);
        }
        else if (Protocol.isStringDescriptor(desc)) {
            if (Protocol.isEnumDescriptor(desc)) {
                simple = false;
                typeName = desc.enum.map(function (str) { return JSON.stringify(str); }).join(' | ');
            }
            else {
                typeName = 'string';
            }
        }
        else if (Protocol.isNumberDescriptor(desc)) {
            typeName = 'number';
        }
        else {
            typeName = desc.type;
        }
        return isArray
            ? simple ? typeName + "[]" : "Array<" + typeName + ">"
            : typeName;
    };
    return ProtocolCodeGenerator;
}());
exports.default = ProtocolCodeGenerator;
function each(arr, cb) {
    if (arr) {
        for (var _i = 0, arr_1 = arr; _i < arr_1.length; _i++) {
            var item = arr_1[_i];
            cb(item);
        }
    }
}
function buildTypeName(name, suffix, domainName) {
    var typeName = name.substring(0, 1).toUpperCase() + name.substring(1) + suffix;
    if (domainName) {
        typeName = domainName + "." + typeName;
    }
    return typeName;
}
